﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Net;

namespace MontoyaTech.Rest.Net
{
    /// <summary>
    /// The outline of a Http Route.
    /// </summary>
    public class Route
    {
        /// <summary>
        /// The http method for this route.
        /// </summary>
        public string Method;

        /// <summary>
        /// The syntax of this route.
        /// </summary>
        public string Syntax;

        /// <summary>
        /// The target function to invoke if this route is invoked.
        /// </summary>
        private Func<RouteContext, HttpListenerResponse> Target;

        /// <summary>
        /// Whether or not to close the response after the route is invoked.
        /// </summary>
        public bool CloseResponse = true;

        internal Route() { }

        /// <summary>
        /// Creates a new route with a given method, syntax, target and optional close response flag.
        /// </summary>
        /// <param name="method"></param>
        /// <param name="syntax"></param>
        /// <param name="target"></param>
        /// <param name="closeResponse"></param>
        public Route(string method, string syntax, Func<RouteContext, HttpListenerResponse> target, bool closeResponse = true)
        {
            this.Method = method;
            this.Syntax = syntax;
            this.Target = target;
            this.CloseResponse = closeResponse;
        }

        /// <summary>
        /// Creates a new route with a given method, syntax, target and optional close response flag.
        /// </summary>
        /// <param name="method"></param>
        /// <param name="syntax"></param>
        /// <param name="target"></param>
        /// <param name="closeResponse"></param>
        public Route(HttpRequestMethod method, string syntax, Func<RouteContext, HttpListenerResponse> target, bool closeResponse = true)
            : this(method.ToString(), syntax, target, closeResponse) { }

        /// <summary>
        /// Invokes this route with a context and a given set of string arguments.
        /// </summary>
        /// <param name="context"></param>
        /// <param name="arguments"></param>
        public virtual void Invoke(RouteContext context, params string[] arguments)
        {
            this.Target.Invoke(context);
        }
    }

    public class Route<T1> : Route
    {
        private Func<RouteContext, T1, HttpListenerResponse> Target;

        public Route(string method, string syntax, Func<RouteContext, T1, HttpListenerResponse> target, bool closeResponse = true)
        {
            this.Method = method;
            this.Syntax = syntax;
            this.Target = target;
            this.CloseResponse = closeResponse;
        }

        public Route(HttpRequestMethod method, string syntax, Func<RouteContext, T1, HttpListenerResponse> target, bool closeResponse = true)
            : this(method.ToString(), syntax, target, closeResponse) { }

        public override void Invoke(RouteContext context, params string[] arguments)
        {
            this.Target.DynamicInvoke(context, RouteArgumentConverter.Convert<T1>(arguments[0]));
        }
    }

    public class Route<T1, T2> : Route
    {
        private Func<RouteContext, T1, T2, HttpListenerResponse> Target;

        public Route(string method, string syntax, Func<RouteContext, T1, T2, HttpListenerResponse> target, bool closeResponse = true)
        {
            this.Method = method;
            this.Syntax = syntax;
            this.Target = target;
            this.CloseResponse = closeResponse;
        }

        public Route(HttpRequestMethod method, string syntax, Func<RouteContext, T1, T2, HttpListenerResponse> target, bool closeResponse = true)
            : this(method.ToString(), syntax, target, closeResponse) { }

        public override void Invoke(RouteContext context, params string[] arguments)
        {
            this.Target.DynamicInvoke(
                context,
                RouteArgumentConverter.Convert<T1>(arguments[0]),
                RouteArgumentConverter.Convert<T2>(arguments[1])
            );
        }
    }

    public class Route<T1, T2, T3> : Route
    {
        private Func<RouteContext, T1, T2, T3, HttpListenerResponse> Target;

        public Route(string method, string syntax, Func<RouteContext, T1, T2, T3, HttpListenerResponse> target, bool closeResponse = true)
        {
            this.Method = method;
            this.Syntax = syntax;
            this.Target = target;
            this.CloseResponse = closeResponse;
        }

        public Route(HttpRequestMethod method, string syntax, Func<RouteContext, T1, T2, T3, HttpListenerResponse> target, bool closeResponse = true)
            : this(method.ToString(), syntax, target, closeResponse) { }

        public override void Invoke(RouteContext context, params string[] arguments)
        {
            this.Target.DynamicInvoke(
                context,
                RouteArgumentConverter.Convert<T1>(arguments[0]),
                RouteArgumentConverter.Convert<T2>(arguments[1]),
                RouteArgumentConverter.Convert<T3>(arguments[2])
            );
        }
    }

    public class Route<T1, T2, T3, T4> : Route
    {
        private Func<RouteContext, T1, T2, T3, T4, HttpListenerResponse> Target;

        public Route(string method, string syntax, Func<RouteContext, T1, T2, T3, T4, HttpListenerResponse> target, bool closeResponse = true)
        {
            this.Method = method;
            this.Syntax = syntax;
            this.Target = target;
            this.CloseResponse = closeResponse;
        }

        public Route(HttpRequestMethod method, string syntax, Func<RouteContext, T1, T2, T3, T4, HttpListenerResponse> target, bool closeResponse = true)
            : this(method.ToString(), syntax, target, closeResponse) { }

        public override void Invoke(RouteContext context, params string[] arguments)
        {
            this.Target.DynamicInvoke(
                context,
                RouteArgumentConverter.Convert<T1>(arguments[0]),
                RouteArgumentConverter.Convert<T2>(arguments[1]),
                RouteArgumentConverter.Convert<T3>(arguments[2]),
                RouteArgumentConverter.Convert<T4>(arguments[3])
            );
        }
    }

    public class Route<T1, T2, T3, T4, T5> : Route
    {
        private Func<RouteContext, T1, T2, T3, T4, T5, HttpListenerResponse> Target;

        public Route(string method, string syntax, Func<RouteContext, T1, T2, T3, T4, T5, HttpListenerResponse> target, bool closeResponse = true)
        {
            this.Method = method;
            this.Syntax = syntax;
            this.Target = target;
            this.CloseResponse = closeResponse;
        }

        public Route(HttpRequestMethod method, string syntax, Func<RouteContext, T1, T2, T3, T4, T5, HttpListenerResponse> target, bool closeResponse = true)
            : this(method.ToString(), syntax, target, closeResponse) { }

        public override void Invoke(RouteContext context, params string[] arguments)
        {
            this.Target.DynamicInvoke(
                context,
                RouteArgumentConverter.Convert<T1>(arguments[0]),
                RouteArgumentConverter.Convert<T2>(arguments[1]),
                RouteArgumentConverter.Convert<T3>(arguments[2]),
                RouteArgumentConverter.Convert<T4>(arguments[3]),
                RouteArgumentConverter.Convert<T4>(arguments[4])
            );
        }
    }

    public class Route<T1, T2, T3, T4, T5, T6> : Route
    {
        private Func<RouteContext, T1, T2, T3, T4, T5, T6, HttpListenerResponse> Target;

        public Route(string method, string syntax, Func<RouteContext, T1, T2, T3, T4, T5, T6, HttpListenerResponse> target, bool closeResponse = true)
        {
            this.Method = method;
            this.Syntax = syntax;
            this.Target = target;
            this.CloseResponse = closeResponse;
        }

        public Route(HttpRequestMethod method, string syntax, Func<RouteContext, T1, T2, T3, T4, T5, T6, HttpListenerResponse> target, bool closeResponse = true)
            : this(method.ToString(), syntax, target, closeResponse) { }

        public override void Invoke(RouteContext context, params string[] arguments)
        {
            this.Target.DynamicInvoke(
                context,
                RouteArgumentConverter.Convert<T1>(arguments[0]),
                RouteArgumentConverter.Convert<T2>(arguments[1]),
                RouteArgumentConverter.Convert<T3>(arguments[2]),
                RouteArgumentConverter.Convert<T4>(arguments[3]),
                RouteArgumentConverter.Convert<T4>(arguments[4]),
                RouteArgumentConverter.Convert<T4>(arguments[5])
            );
        }
    }

    public class Route<T1, T2, T3, T4, T5, T6, T7> : Route
    {
        private Func<RouteContext, T1, T2, T3, T4, T5, T6, T7, HttpListenerResponse> Target;

        public Route(string method, string syntax, Func<RouteContext, T1, T2, T3, T4, T5, T6, T7, HttpListenerResponse> target, bool closeResponse = true)
        {
            this.Method = method;
            this.Syntax = syntax;
            this.Target = target;
            this.CloseResponse = closeResponse;
        }

        public Route(HttpRequestMethod method, string syntax, Func<RouteContext, T1, T2, T3, T4, T5, T6, T7, HttpListenerResponse> target, bool closeResponse = true)
            : this(method.ToString(), syntax, target, closeResponse) { }

        public override void Invoke(RouteContext context, params string[] arguments)
        {
            this.Target.DynamicInvoke(
                context,
                RouteArgumentConverter.Convert<T1>(arguments[0]),
                RouteArgumentConverter.Convert<T2>(arguments[1]),
                RouteArgumentConverter.Convert<T3>(arguments[2]),
                RouteArgumentConverter.Convert<T4>(arguments[3]),
                RouteArgumentConverter.Convert<T4>(arguments[4]),
                RouteArgumentConverter.Convert<T4>(arguments[5]),
                RouteArgumentConverter.Convert<T4>(arguments[6])
            );
        }
    }

    public class Route<T1, T2, T3, T4, T5, T6, T7, T8> : Route
    {
        private Func<RouteContext, T1, T2, T3, T4, T5, T6, T7, T8, HttpListenerResponse> Target;

        public Route(string method, string syntax, Func<RouteContext, T1, T2, T3, T4, T5, T6, T7, T8, HttpListenerResponse> target, bool closeResponse = true)
        {
            this.Method = method;
            this.Syntax = syntax;
            this.Target = target;
            this.CloseResponse = closeResponse;
        }

        public Route(HttpRequestMethod method, string syntax, Func<RouteContext, T1, T2, T3, T4, T5, T6, T7, T8, HttpListenerResponse> target, bool closeResponse = true)
            : this(method.ToString(), syntax, target, closeResponse) { }

        public override void Invoke(RouteContext context, params string[] arguments)
        {
            this.Target.DynamicInvoke(
                context,
                RouteArgumentConverter.Convert<T1>(arguments[0]),
                RouteArgumentConverter.Convert<T2>(arguments[1]),
                RouteArgumentConverter.Convert<T3>(arguments[2]),
                RouteArgumentConverter.Convert<T4>(arguments[3]),
                RouteArgumentConverter.Convert<T4>(arguments[4]),
                RouteArgumentConverter.Convert<T4>(arguments[5]),
                RouteArgumentConverter.Convert<T4>(arguments[6]),
                RouteArgumentConverter.Convert<T4>(arguments[7])
            );
        }
    }
}
